#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <vector>
#include "fitsio.h"
#include "psrfits.h"
#include "guppi_params.h"
#include "fitshead.h"
#include "median.h"
#include "setimysql.h"
#include <fftw3.h>
#include <sys/stat.h>
#include "barycenter.h"
#include "rawdopplersearch.h"
#include "setilib.h"
#include "mb_read_blocks_guppi.h"
#include "mb_splitter.h"
#include "mb_coords.h"

//#include "seti_dr2utils.h"

/* Guppi channel-frequency mapping */
/* sample at 1600 MSamp for an 800 MHz BW */
/* N = 256 channels with frequency centers at m * fs/N */
/* m = 0,1,2,3,... 255 */

// guppi raw data layout (data portion of a ~psrfits HDU (subintegration)):
//
// First of 32 frequency channels / subintegration:
// Pol0Samp0    Pol0Samp0   Pol1Samp0   Pol1Samp0   ... Pol1Samp255 Pol1Samp255
// real         imag        real        imag            real        imag
// |--------------------------------------------|       ----------------------|
//   byte 0 (2 bit data, quantized from 8 bit)              ... byte 255
//   time 0                                                 ... time 255
//
// The next subintegration will be the same 32 channels, starting with time 256.
// Thus, each subintegration is 32*256 = 8192 bytes of data.
//
#if 0

struct gpu_input {
    char *file_prefix;
    struct                  guppi_params gf;
    struct                  psrfits pf;
    unsigned int filecnt;
    FILE *fil;
    int invalid;
    int curfile;
    int overlap;   /* add this keyword here since it doesn't seem to appear in guppi_params.c */
    long int first_file_skip; /* in case there's 8bit data in the header of file 0 */
};

struct guppi_params {
    /* Packet information for the current block */
    long long packetindex;      // Index of first packet in raw data block
    int packetsize;             // Size in bytes of data portion of each packet
    int n_packets;              // Total number of packets in current block
    int n_dropped;              // Number of packets dropped in current block
};

struct psrfits {
    char mode;              // Read (r) or write (w).
    struct                  hdrinfo hdr;
    struct                  subint sub;
};

struct hdrinfo {
    char source[24];        // Source name
    char frontend[24];      // Frontend used
    long double MJD_epoch;  // Starting epoch in MJD
    double fctr;            // Center frequency of the observing band (MHz)
    double df;              // Frequency spacing between the channels (MHz)
    double BW;              // Bandwidth of the observing band (MHz)
    int nbits;              // Number of bits per data sample
    int nchan;              // Number of channels
    int rcvr_polns;         // Number of polns provided by the receiver
};

struct subint {
    double tsubint;         // Length of subintegration (sec)
    double offs;            // Offset from Start of subint centre (sec)
    double ra;              // RA (J2000) at subint centre (deg)
    double dec;             // Dec (J2000) at subint centre (deg)
    int bytes_per_subint;   // Number of bytes for one row of raw data
    unsigned char *data;    // Ptr to the raw data itself
};

#endif

#define RAW_DATA_HEADER_BUF_SIZE 32768

typedef struct {
    // file control...
    long long int   startindx;
    long long int   curindx;
	long long int   chanbytes;
	long long int   subint_offset;
	long int        currentblock;
	int             indxstep;
    // user prefs...
    int             channel;
    int             polarization;
    long int        startblock; 
    long int        numblocks; 
    int             vflag; 
} control_block_t;

/* Parse info from buffer into param struct */
extern void guppi_read_obs_params(char *buf, 
                                     struct guppi_params *g,
                                     struct psrfits *p);

/* prototypes */

int exists(const char *fname);

int unpack_samples(unsigned char * raw, long int count, int pol, dr2_compact_block_t &this_tapebuffer);

//-------------------------------------------------------
int print_header(dr2_compact_block_t &this_tapebuffer) {
//-------------------------------------------------------

	fprintf(stderr, "setiheader : Name: %s Header Size: %d Data Size: %d Channel: %d Polarization: N samplerate %lf sky_freq %ld Data Time: %lf Coordinate Time: %lf RA: %f Dec: %f\n", 
            this_tapebuffer.header.name,
            this_tapebuffer.header.header_size,
	        this_tapebuffer.header.data_size,		
	        this_tapebuffer.header.channel,		
	        this_tapebuffer.header.samplerate,		
	        this_tapebuffer.header.sky_freq,		
            this_tapebuffer.header.data_time.mjd().uval(),
            this_tapebuffer.header.coord_time.mjd().uval(),
	        this_tapebuffer.header.ra,
	        this_tapebuffer.header.dec);
}

//-------------------------------------------------------
int populate_header(
	            control_block_t     &control_block,
                struct gpu_input    &rawinput,
                dr2_compact_block_t &this_tapebuffer
) {
//-------------------------------------------------------

    /* size of this header */
    this_tapebuffer.header.header_size = sizeof(this_tapebuffer.header);

    /* number of samples * size of a signed char * 2 for real/imag */
    this_tapebuffer.header.data_size = (control_block.chanbytes * sizeof(signed char) * 2);

    /* populate with source name from psrfits header */
    sprintf(this_tapebuffer.header.name, "%s", rawinput.pf.hdr.source);

    /* populate channel and polarization */
    this_tapebuffer.header.channel      = control_block.channel;
    //this_tapebuffer.header.polarization = control_block.polarization;   // TODO add polarization

    /* (complex) sample rate in Hz */
    this_tapebuffer.header.samplerate = rawinput.pf.hdr.df * 1000000;

    this_tapebuffer.header.sky_freq =   (long int) (
                              (double) 1000000 * 
                              ( 
                                (rawinput.pf.hdr.fctr   - 
                                (rawinput.pf.hdr.BW/2)) + 
                                ((control_block.channel+0.5) * rawinput.pf.hdr.df)
                              )
                            );

    print_header(this_tapebuffer);
}


//-------------------------------------------------------
int get_params(
	            control_block_t     &control_block,
                struct gpu_input    &rawinput,
                char                header_buf[],
                char                filname[],
	            unsigned char *     &channelbuffer	
) {
//-------------------------------------------------------
	long long int chanbytes_overlap = 0;
        // Read first raw data header
	    if(fread(header_buf, sizeof(char), RAW_DATA_HEADER_BUF_SIZE, rawinput.fil) == RAW_DATA_HEADER_BUF_SIZE){
		   
		    guppi_read_obs_params(header_buf, &rawinput.gf, &rawinput.pf);
		   
            // Get to first 2 bit data header...
            // At regular intervals, as a sanity check, the 2 bit
            // data are preceeded by the full 8 bit version of
            // the same data.  We ignore the 8 bit data here.
		    if(rawinput.pf.hdr.nbits == 8) {
			    fprintf(stderr, "caught an 8 bit header... moving on to first/next 2 bit header\n");
			    /* figure out the size of the first subint + header */
		        rawinput.first_file_skip = rawinput.pf.sub.bytes_per_subint + gethlength(header_buf);
			    /* rewind to the beginning */	
		   	    fseek(rawinput.fil, -RAW_DATA_HEADER_BUF_SIZE, SEEK_CUR);
		   	    /* seek past the first subint + header */
		   	    fseek(rawinput.fil, rawinput.first_file_skip, SEEK_CUR);
			    /* read the next header */
		   	    fread(header_buf, sizeof(char), RAW_DATA_HEADER_BUF_SIZE, rawinput.fil);
			    guppi_read_obs_params(header_buf, &rawinput.gf, &rawinput.pf);
			    fclose(rawinput.fil);
		    } else {
                // not 8 bit data 
			    fclose(rawinput.fil);
		    }   // end if(rawinput.pf.hdr.nbits == 8)

		    /* we'll use this header to set the params for the whole observation */
		   
		    rawinput.fil = NULL;

		    hgeti4(header_buf, "OVERLAP", &rawinput.overlap);
			
		    fprintf(stderr, "   packetindex       : %lld\n", rawinput.gf.packetindex);
		    fprintf(stderr, "   packetsize        : %d\n", rawinput.gf.packetsize);
		    fprintf(stderr, "   n_packets         : %d\n", rawinput.gf.n_packets);
		    fprintf(stderr, "   n_dropped         : %d\n",rawinput.gf.n_dropped);
		    fprintf(stderr, "   bytes_per_subint  : %d\n",rawinput.pf.sub.bytes_per_subint);
		    fprintf(stderr, "   overlap           : %d\n",rawinput.overlap);
	        } else {    
	  		    fprintf(stderr, "couldn't read a header\n");
			    return 0;
	        }   // end fread(buf...


    if(control_block.vflag>=1) fprintf(stderr, "calculating index step\n");

    /* number of packets that we *should* increment by */
    control_block.indxstep = (int) ((rawinput.pf.sub.bytes_per_subint * 4) / rawinput.gf.packetsize)              - 
                      (int) (rawinput.overlap * rawinput.pf.hdr.nchan * rawinput.pf.hdr.rcvr_polns  * 
                      2 / rawinput.gf.packetsize);

    //spectraperint = control_block.indxstep * band[first_good_band].gf.packetsize / (band[first_good_band].pf.hdr.nchan * band[first_good_band].pf.hdr.rcvr_polns * 2 * ftacc);
    //spectraperint = ((rawinput.pf.sub.bytes_per_subint / rawinput.pf.hdr.nchan) - rawinput.overlap) / ftacc;	

    //overlap = rawinput.overlap;

    if (control_block.channel >= rawinput.pf.hdr.nchan) {
	    fprintf(stderr, "control_block.channel %d more than channels in data %d\n", control_block.channel, rawinput.pf.hdr.nchan);
	    return 0;
    } else {
	    fprintf(stderr, "Numer of channels in file %d\n", rawinput.pf.hdr.nchan);
    }

    /* number of non-overlapping bytes in each channel */
    /* control_block.indxstep increments by the number of unique packets in each sub-integration */
    /* packetsize is computed based on the original 8 bit resolution */
    /* divide by 4 to get to 2 bits, nchan to get to number of channels */

    control_block.chanbytes = control_block.indxstep * rawinput.gf.packetsize / (4 * rawinput.pf.hdr.nchan); 
    fprintf(stderr, "chan bytes %lld\n", control_block.chanbytes);

    // commented out by jeffc - channelbuffer is unused 
    //channelbuffer  = (unsigned char *) calloc( control_block.chanbytes , sizeof(char) );
    //if (!channelbuffer) {
	//    fprintf(stderr, "error: couldn't allocate memory for buffer\n");
	//    return 0;
    //}

    //fprintf(stderr, "malloc'ed %Ld bytes for channel %d\n",  control_block.chanbytes, control_block.channel );	

    //	tstart=band[first_good_band].pf.hdr.MJD_epoch;
    //	tsamp = band[first_good_band].pf.hdr.dt * ftacc;

    //	strcat(buf, strtok(band[first_good_band].pf.hdr.ra_str, ":"));
    //	strcat(buf, strtok(band[first_good_band].pf.hdr.dec_str, ":"));

    /*
    fprintf(stderr, "freq %ld\n", setiheader.sky_freq);
    fprintf(stderr, "fctr %lf\n", rawinput.pf.hdr.fctr);
    fprintf(stderr, "bandwidth %lf\n", rawinput.pf.hdr.BW);
    fprintf(stderr, "freqdouble %lf\n", ((double) 1000000 * ((rawinput.pf.hdr.fctr - (rawinput.pf.hdr.BW/2)) + ((control_block.channel+0.5) * rawinput.pf.hdr.df))));
    */

    /* total number of bytes per channel, including overlap */
    chanbytes_overlap = rawinput.pf.sub.bytes_per_subint / rawinput.pf.hdr.nchan;

    /* memory offset for our chosen channel within a subint */
    control_block.subint_offset = control_block.channel * chanbytes_overlap;

    if(control_block.vflag>=1) fprintf(stderr, "Index step: %d\n", control_block.indxstep);
    if(control_block.vflag>=1) fprintf(stderr, "bytes per subint %d\n",rawinput.pf.sub.bytes_per_subint );
}


//-------------------------------------------------------
int read_block(
	            control_block_t     &control_block,
                struct gpu_input    &rawinput,
                char                header_buf[],
                char                filname[],
                double              &start_data_time,
                double              &end_data_time,
                long int            file_cnt,
                dr2_compact_block_t &this_tapebuffer
) {
//-------------------------------------------------------

	size_t rv=0;
    int retval=0;
	//long int j=0;

	    if(!rawinput.invalid){						  
		    if(rawinput.fil == NULL) {
			    /* no file is open for this band, try to open one */
			    sprintf(filname, "%s.%04d.raw",rawinput.file_prefix,rawinput.curfile);
			    if(exists(filname)){
                    fprintf(stderr, "Opening %s\n", filname);
				    rawinput.fil = fopen(filname, "rb");			 
				    if(rawinput.curfile == 0 && rawinput.first_file_skip != 0) fseek(rawinput.fil, rawinput.first_file_skip, SEEK_CUR);  
			    }	else {
			  	    rawinput.invalid = 1;
		  	  	    fprintf(stderr, "couldn't open any more files!\n");
		  	    }
		    }

		    if(rawinput.fil){		  
		        fprintf(stderr, "Reading first/next 32KB\n"); 
	            //if((fread(buf, sizeof(char), RAW_DATA_HEADER_BUF_SIZE, rawinput.fil) == RAW_DATA_HEADER_BUF_SIZE) && (control_block.currentblock < (control_block.numblocks + control_block.startblock - 1)) ) {				
	            if(fread(header_buf, sizeof(char), RAW_DATA_HEADER_BUF_SIZE, rawinput.fil) == RAW_DATA_HEADER_BUF_SIZE) {
		            fprintf(stderr, "...success!\n"); 
					fseek(rawinput.fil, -RAW_DATA_HEADER_BUF_SIZE, SEEK_CUR);                                  // reposition to the start of this 32KB
					if(control_block.vflag>=1) fprintf(stderr, "header length: %d\n", gethlength(header_buf));
					guppi_read_obs_params(header_buf, &rawinput.gf, &rawinput.pf);

					control_block.currentblock = (long int) ((double) rawinput.gf.packetindex/ (double) control_block.indxstep);

					if(control_block.vflag>=1) {
						 fprintf(stderr, "packetindex %Ld packetsize: %d n_packets %d n_dropped: %d blocks: %ld RA: %f DEC: %f subintoffset %f tsubint %f MJD %Lf\n", 
                                    rawinput.gf.packetindex,
						            rawinput.gf.packetsize,
						            rawinput.gf.n_packets,
						            rawinput.gf.n_dropped,
						            control_block.currentblock,
						            rawinput.pf.sub.ra,
						            rawinput.pf.sub.dec,
						            rawinput.pf.sub.offs,
						            rawinput.pf.sub.tsubint,
						            rawinput.pf.hdr.MJD_epoch);
					}
					
					/* populate the variables that change with each block */
				    /* RA (J2000) at subint centre (deg), Dec (J2000) at subint centre (deg), time in MJD */
					this_tapebuffer.header.ra           = rawinput.pf.sub.ra/15.0;      // splitter expects RA in decimal hours            
					this_tapebuffer.header.dec          = rawinput.pf.sub.dec;           
					/* increment time for data by 0.5 x length of block to push quoted time to the _last_ sample in the block */					
					double data_mjd                           = rawinput.pf.hdr.MJD_epoch + ((rawinput.pf.sub.offs + rawinput.pf.sub.tsubint/2)/86400.0);     
					double coord_mjd                          = rawinput.pf.hdr.MJD_epoch + (rawinput.pf.sub.offs/86400.0);    
                    this_tapebuffer.header.data_time    = seti_time(days(data_mjd),MJD0);
                    this_tapebuffer.header.coord_time   = seti_time(days(coord_mjd),MJD0);

                    if(control_block.currentblock == 1) {
                        start_data_time = this_tapebuffer.header.data_time.mjd().uval();
                    } else {
                        end_data_time = this_tapebuffer.header.data_time.mjd().uval();     // assigned many times, will end up with the final block's time
                    }
					
			   		if(rawinput.gf.packetindex == control_block.curindx) {
						 /* read a subint with correct index */

						 fseek(rawinput.fil, gethlength(header_buf), SEEK_CUR);
						 rv=0;
						 
						 if (control_block.currentblock > (control_block.startblock-1)){
						 
						 	rv=fread(rawinput.pf.sub.data, sizeof(char), rawinput.pf.sub.bytes_per_subint, rawinput.fil);  // read raw samples		 
						 	
							if( ((long int)rv == rawinput.pf.sub.bytes_per_subint) ){
							   if(control_block.vflag>=1) fprintf(stderr,"read %d bytes from %ld in curfile %d\n", 
                                                                  rawinput.pf.sub.bytes_per_subint, file_cnt, rawinput.curfile);						
							   unpack_samples(rawinput.pf.sub.data + control_block.subint_offset, 
                                              control_block.chanbytes, 
                                              control_block.polarization,
                                              this_tapebuffer);   					               // unpack samples to the return vector   

                               // The following two lines were commented out in Andrews code.
							   //memcpy(channelbuffer + channelbuffer_pos, rawinput.pf.sub.data + control_block.subint_offset, control_block.chanbytes);										   
							   //channelbuffer_pos = channelbuffer_pos + control_block.chanbytes;

							} else {
								rawinput.fil = NULL;
								rawinput.invalid = 1;
								fprintf(stderr,"ERR: couldn't read as much as the header said we could... assuming corruption and exiting...\n");
								return 0;
							}

						 } else (rv = fseek(rawinput.fil, rawinput.pf.sub.bytes_per_subint, SEEK_CUR)); 
						 
					} else if( (rawinput.gf.packetindex > control_block.curindx) && (control_block.currentblock > (control_block.startblock-1)) ) {

						 fprintf(stderr,"ERR: control_block.curindx: %Ld, pktindx: %Ld Did we drop a whole subintegration? Reusing last subint...\n", 
                                 control_block.curindx, rawinput.gf.packetindex );
						 /* read a subint with too high an indx, must have dropped a whole subintegration*/

						/* pf.sub.data *should* still contain the last valid subint */
						/* grab a copy of the last subint  - probably should add gaussian noise here, but this is better than nothing! */

						/* we'll keep the ra/dec values from this subint, but push back the time to keep everything sensible */
					    data_mjd = rawinput.pf.hdr.MJD_epoch + ((rawinput.pf.sub.offs - rawinput.pf.sub.tsubint/2)/86400.0) ;     
                        this_tapebuffer.header.data_time = seti_time(days(data_mjd),MJD0);

                        // TODO keep the following demo buffer code to remind me what to do here
                        // header to return buffer
						//memcpy((*setibuffer) + control_block.setibuffer_pos, 
                        //       &setiheader, 
                        //       sizeof(setiheader));
						//control_block.setibuffer_pos += sizeof(setiheader);

                        // data to return buffer
						//memmove((*setibuffer) + control_block.setibuffer_pos, 
                        //        (*setibuffer) + control_block.setibuffer_pos - sizeof(setiheader) - control_block.chanbytes, 
                        //        (control_block.chanbytes * 2 * sizeof(unsigned char)));
						//control_block.setibuffer_pos += (control_block.chanbytes * 2 * sizeof(unsigned char));

						/* We'll get the current valid subintegration again on the next time through this loop */

					} else if(rawinput.gf.packetindex < control_block.curindx) {
						 fprintf(stderr,"Error expecting a higher packet index than we got control_block.curindx: %Ld, pktindx: %Ld\n", 
                                 control_block.curindx, rawinput.gf.packetindex );
						 /* somehow we were expecting a higher packet index than we got !?!? */	
						 fprintf(stderr, "assuming corruption and exiting...\n");
						 return(0);
					}
    
                if(control_block.vflag>=1) fprintf(stderr, "Read block %ld\n", control_block.currentblock);
                //if(control_block.vflag>=1) {print_header_old(setiheader); print_header(this_tapebuffer);}
                if(control_block.vflag>=1) print_header(this_tapebuffer);
                retval = true;
				} else {
				   fprintf(stderr, "Could not read first/next 32KB\n"); 
				   fclose(rawinput.fil);
				   rawinput.fil = NULL;
				   rawinput.curfile++;						
                   retval = false;
				}   // end if((fread(buf...)
		    }   // end if(rawinput.fil)			 	 	 
	    }   // end if(!rawinput.invalid)
										
	    if(rawinput.fil != NULL) {
            control_block.curindx = control_block.curindx + control_block.indxstep;
            retval = 1;     // we have a new block
        } else {
            retval = 0;     // we do not have a new block
        }

    return(retval);
}

//-------------------------------------------------------
long int read_blocks_guppi(char *           file_prefix,
                           long int         startblock, 
                           long int         numblocks, 
                           int              channel, 
                           int              polarization, 
                           int              vflag 
) {
//-------------------------------------------------------

	int overlap; // amount of overlap between sub integrations in samples

    int retval=0;

	int filecnt=0;
    char header_buf[RAW_DATA_HEADER_BUF_SIZE];   // we read the raw data headers into this buffer

	unsigned char *channelbuffer;	

	char filname[250];

    dr2_compact_block_t this_tapebuffer;

	struct gpu_input rawinput;	

    control_block_t control_block;
    control_block.chanbytes     = 0;
    control_block.subint_offset = 0;
    control_block.currentblock  = 0;
    control_block.indxstep      = 0;
    control_block.channel       = channel;
    control_block.polarization  = polarization;
    control_block.numblocks     = numblocks;
    control_block.startblock    = startblock;
    control_block.vflag         = vflag;

    size_t rv=0;
	long unsigned int by=0;
	long int i=0,j=0;

    double start_data_time;
    double end_data_time;

	rawinput.file_prefix = NULL;
	rawinput.fil = NULL;
	rawinput.invalid = 0;
	rawinput.first_file_skip = 0;  

    opterr = 0;
    rawinput.file_prefix=file_prefix;
     
    /* no input specified */
    if(rawinput.file_prefix == NULL) {
	    printf("WARNING no input stem specified%ld\n", (i+1));
	    return 0;
    }

    if(strstr(rawinput.file_prefix, ".0000.raw") != NULL) memset(rawinput.file_prefix + strlen(rawinput.file_prefix) - 9, 0x0, 9);

    // get size of data set, in terms of file count and byte count
    fprintf(stderr, "Finding size of data set...\n");
    j = 0;
    struct stat st;
    long int size=0;
    do {
	    sprintf(filname, "%s.%04ld.raw",rawinput.file_prefix,j);
	    fprintf(stderr, "  trying %s ...",filname);		
	    j++;
	    if(exists(filname)) { 
            fprintf(stderr, " found\n");
		    stat(filname, &st);
		    size = size + st.st_size;
	    } else {
            fprintf(stderr, " not found\n");
        }
    } while (exists(filname));
    rawinput.filecnt = j-1;
    fprintf(stderr, "  File count is %i  Total size is %ld bytes\n",rawinput.filecnt, size);

    /* didn't find any files */
    if(rawinput.filecnt < 1) {
	    fprintf(stderr, "no files for stem %s found\n",rawinput.file_prefix);
	    exit(1);		
    }

    /* open the first file for input */
    sprintf(filname, "%s.0000.raw", rawinput.file_prefix);
    fprintf(stderr, "Opening %s\n", filname);
    rawinput.fil = fopen(filname, "rb");
    if(!rawinput.fil){
        fprintf(stderr, "couldn't open first file\n");
	    return 0;
    }  

    get_params(
                control_block,
                rawinput,
                header_buf,
                filname,
                channelbuffer
    );

    // Now that we know our sizing requirements, allocate the buffers...
    // ...raw data read buffer (allocate and free each time because we don 't know 
    // how many times we'll be here)
    rawinput.pf.sub.data  = (unsigned char *) malloc(rawinput.pf.sub.bytes_per_subint);
    if(!rawinput.pf.sub.data) {
	    fprintf(stderr, "error: couldn't allocate memory for the raw data read buffer\n");
	    return 0;
    } else {
        fprintf(stderr, "malloc'ed %ld bytes at %p for the raw data read buffer!\n", rawinput.pf.sub.bytes_per_subint, rawinput.pf.sub.data);
    }

    // populate the header with static items
    populate_header(
	            control_block,
                rawinput,
                this_tapebuffer
    );

    //fflush(stdout);

    control_block.startindx        = rawinput.gf.packetindex;
    control_block.curindx          = control_block.startindx;
    filecnt                        = rawinput.filecnt;
    rawinput.curfile               = 0;			

    // Read blocks until we run out of data or the numblocks request is met
    do{    
        retval = read_block(
	                control_block,
                    rawinput,
                    header_buf,
                    filname,
                    start_data_time,
                    end_data_time,
                    j,
                    this_tapebuffer
                 );
        if(retval) {
            //tapebuffer.push_back(this_tapebuffer);
            tapebuffer.push_back(this_tapebuffer);
            this_tapebuffer.data.clear();
        }
    } while((!(rawinput.invalid)) && (control_block.currentblock <= (numblocks + control_block.startblock - 1)));

    fprintf(stderr, "getting coordinate history...\n");
    get_coord_history();
	
    fprintf(stderr, "finishing up...\n");

    if(control_block.vflag>=1) fprintf(stderr, "bytes: %ld\n",by);
	
    // free the raw data read buffer
    if (rawinput.pf.sub.data) {
        fprintf(stderr, "freeing rawinput.pf.sub.data at %p\n", rawinput.pf.sub.data);
	    free(rawinput.pf.sub.data);
        rawinput.pf.sub.data = NULL;
	    fprintf(stderr, "freed subint data buffer\n");
    }

    fprintf(stderr, "done\n");

    // commented out by jeffc
    //free(channelbuffer);

    int blocks_read = tapebuffer.size();
    if(control_block.vflag>=1) fprintf(stderr, "grabbed %Ld blocks covering %lf seconds of time\n", 
                                       blocks_read, (end_data_time-start_data_time)*86400);
    return blocks_read;

}   // end read_blocks_guppi()

//-------------------------------------------------------
int unpack_samples(unsigned char * raw, long int count, int pol, dr2_compact_block_t &this_tapebuffer) {
//-------------------------------------------------------
// unpack guppi 2 bit complex samples into a vector of complex signed chars
// pol is polarization (0 for X, 1 for Y) 

	 int i;
     int shift_factor = pol * 2;    // shift_factor will be 0 or 2
	 float quantlookup[4];
     std::complex<signed char> sample;

     // the * 10 is to keep some of the precision intact as we
     // go to signed char  
	 quantlookup[0] = 3.3358750     * 10;
	 quantlookup[1] = 1.0           * 10;      
	 quantlookup[2] = -1.0          * 10;
	 quantlookup[3] = -3.3358750    * 10;

	 for(i=0; i < count; i++) {
          // real (2 bits of raw data)
		  sample.real( (signed char)quantlookup[( raw[i] >> (shift_factor * 2)      & 1)        +   // bit 0 or 4  
                                                ((raw[i] >> (shift_factor * 2 + 1)  & 1) * 2)]      // bit 1 or 5
                     ); 
		  // imag (2 bits of raw data)
		  sample.imag( (signed char)quantlookup[( raw[i] >> (shift_factor+1 * 2)      & 1)      +   // bit 2 or 6 
                                                ((raw[i] >> (shift_factor+1 * 2 + 1)  & 1) * 2)]    // bit 3 or 7
                     ); 
          // add this sample to our vector
          this_tapebuffer.data.push_back(sample);
	 }

//fprintf(stderr, "old count : %ld    new count : %ld\n", count,  this_tapebuffer.data.size());
//fprintf(stderr, "old data  : %d %d  new data  :  %d %d\n", samples[0], samples[1], this_tapebuffer.data[0].real(), this_tapebuffer.data[0].imag());
//fprintf(stderr, "old data  : %d %d  new data  :  %d %d\n", samples[2], samples[3], this_tapebuffer.data[1].real(), this_tapebuffer.data[1].imag());
//fprintf(stderr, "old data  : %d %d  new data  :  %d %d\n", samples[count*2-2], samples[count*2-1], this_tapebuffer.data[count-1].real(), this_tapebuffer.data[count-1].imag());

	 return 0;
}

//-------------------------------------------------------
int exists(const char *fname) {
//-------------------------------------------------------
    FILE *file;
    if ((file = fopen(fname, "r"))) {
        fclose(file);
        return 1;
    }
    return 0;
}
